(in-package "ACL2")

; This file is mostly about first-order terms, formulas, and
; substitution of terms for variables.
;
; Author:  William McCune (mccune@mcs.anl.gov).

(defmacro natp (n)
    (list 'and (list 'integerp n) (list '<= 0 n)))

; Reserved logic symbols.  This is mostly to keep users from
; writing confusing formulas.

(defun logic-symbolp (x)
  (declare (xargs :guard t))
  (or (equal x 'and)
      (equal x 'or)
      (equal x 'not)
      (equal x 'imp)
      (equal x 'iff)
      (equal x 'all)
      (equal x 'exists)))

; There are 2 kinds of terms and formulas: ordinary and -d.
;
; The -d kind can (but don't have to) contain "domain-term"s.
; A domain-term is an element of the domain of some interpretation.
; When a formula is being evaluated in an interpretation, we can
; substitute domain-terms for free variables.
;
; The functions that operate on formulas have guards saying that
; the formulas are wff-d.
;
; Ordinary terms and formulas:
;
; wft: well-formed term
; wfa: well-formed atomic formula
; wff: well-formed formula
;
; Terms and formulas that can contain domain elements:
;
; wft-d: well-formed domain term
; wfa-d: well-formed domain atomic formula
; wff-d: well-formed domain formula
;
; To prevent duplicated code, it might be better to have just one
; set of functions which take an argument asking whether the thing
; is ordinary or -d.
;
; Or, it might be better to allow all formulas and terms to contain
; domain terms.  When defining a new function on terms or formulas,
; it's not always clear if domain-terms should be accomodated.
; 
;=================================================================
; Well-formed terms

; Domain terms are special terms that are elements of the domain
; of an interpretation.  Let them be naturals.

(defun domain-term (x)
  (declare (xargs :guard t))
  (natp x))

; A variable is either a symbolp or a (symbolp . natp).
; The pairs are convenient for introducing new variables.

(defun variable-term (x)
  (declare (xargs :guard t))
  (or (symbolp x) (and (consp x)
		       (symbolp (car x))
		       (natp (cdr x)))))

; A function symbol can be either a symbolp or a (symbolp . natp).
; The pairs are used as skolem functions.  Constants are functions
; with no arguments, so this type includes constant symbols as well.
		
(defun function-symbol (x)
  (declare (xargs :guard t))
  (or (symbolp x) (and (consp x)
		       (symbolp (car x))
		       (natp (cdr x)))))

; A relation symbol is a symbolp that is not a logic-symbolp.
; This means that a relation symbol is also a function symbol.
; Atomic formulas are distinguished from terms by context.

(defun relation-symbol (x)
  (declare (xargs :guard t))
  (and (symbolp x) (not (logic-symbolp x))))

(defun wft-list (l)
  (declare (xargs :guard t))
  (if (not (true-listp l))
      nil
      (if (null l)
	  t
	  (and (wft-list (cdr l))
	       (if (variable-term (car l))
		   t
		   (and (true-listp (car l))
			(consp (car l))
			(function-symbol (caar l))
			(wft-list (cdar l))))))))

(defun wft (x)
  (declare (xargs :guard t))
  (wft-list (list x)))

(defun wft-d-list (l)
  (declare (xargs :guard t))
  (if (not (true-listp l))
      nil
      (if (null l)
	  t
	(and (wft-d-list (cdr l))
	     (if (or (variable-term (car l)) (domain-term (car l)))
		 t
	         (and (true-listp (car l))
		      (consp (car l))
		      (function-symbol (caar l))
		      (wft-d-list (cdar l))))))))

(defun wft-d (x)
  (declare (xargs :guard t))
  (wft-d-list (list x)))

(defthm wft-list-is-wft-d-list
  (implies (wft-list l)
	   (wft-d-list l)))

;=================================================================
; Well-formed formulas.
;
; Functions to get the first and second arguments of a formula.

(defmacro a1 (p)  ; argument 1
  (list 'cadr p))

(defmacro a2 (p) ; argument 2
  (list 'caddr p))

; These functions check for true lists of length 2 and 3.
; I don't know if it would be better to use true-listp and length.

(defun list2p (l)
  (declare (xargs :guard t))
  (and (consp l) (consp (cdr l)) (null (cddr l))))

(defun list3p (l)
  (declare (xargs :guard t))
  (and (consp l) (list2p (cdr l))))

; These functions check if a formula has a particular type and
; is well-formed at the top level.

(defun wfand (p)
  (declare (xargs :guard t))
  (and (list3p p) (equal (car p) 'and)))

(defun wfor  (p)
  (declare (xargs :guard t))
  (and (list3p p) (equal (car p) 'or)))

(defun wfnot (p)
  (declare (xargs :guard t))
  (and (list2p p) (equal (car p) 'not)))

(defun wfimp (p)
  (declare (xargs :guard t))
  (and (list3p p) (equal (car p) 'imp)))

(defun wfiff (p)
  (declare (xargs :guard t))
  (and (list3p p) (equal (car p) 'iff)))

(defun wfall (p)
  (declare (xargs :guard t))
  (and (list3p p) (equal (car p) 'all) (variable-term (a1 p))))

(defun wfexists (p)
  (declare (xargs :guard t))
  (and (list3p p) (equal (car p) 'exists) (variable-term (a1 p))))

; Well-formed atomic formula (does not include 'true or 'false).
; Actually, because relation-symbol and function-symbol can overlap,
; nonvariable terms can also be recognized by these
; functions.  What these really check is that a term is not a
; complex formula (that is, and, or, not, imp, iff, all, exists).
; So if the context is a formula, and wfa is satisfied, then the term is
; an atomic formula.

(defun wfa (a)
  (declare (xargs :guard t))
  (and (true-listp a)
       (consp a)
       (relation-symbol (car a))
       (wft-list (cdr a))))

(defun wfa-d (a)
  (declare (xargs :guard t))
  (and (true-listp a)
       (consp a)
       (relation-symbol (car a))
       (wft-d-list (cdr a))))

; Well-formed formula.

(defun wff (f)
  (declare (xargs :guard t))
  (cond ((equal f 'true)  t)
	((equal f 'false) t)
	((wfa f) t)
	((wfnot f) (wff (a1 f)))
	((or (wfand f)
	     (wfor  f)
	     (wfimp f)
	     (wfiff f)) (and (wff (a1 f)) (wff (a2 f))))
	((or (wfall f)
	     (wfexists f)) (wff (a2 f)))
	(t nil)))
  
(defun wff-d (f)
  (declare (xargs :guard t))
  (cond ((equal f 'true) t)
	((equal f 'false) t)
	((wfa-d f) t)
	((wfnot f) (wff-d (a1 f)))
	((or (wfand f)
	     (wfor  f)
	     (wfimp f)
	     (wfiff f)) (and (wff-d (a1 f)) (wff-d (a2 f))))
	((or (wfall f)
	     (wfexists f)) (wff-d (a2 f)))
	(t nil)))

; The following really slows down proofs, and it can usually be
; proved on the fly.
;  
; (defthm wff-is-wff-d
;   (implies (wff a)
;	   (wff-d a)))

; Simple formula: well-formed atomic formula, 'true, or 'false.

(defun simple-formula (f)
  (declare (xargs :guard t))
  (or (equal f 'true)
      (equal f 'false)
      (wfa f)))

(defun simple-formula-d (f)
  (declare (xargs :guard t))
  (or (equal f 'true)
      (equal f 'false)
      (wfa-d f)))

;=========================================================
;
; Substitute terms in terms and formulas.
;
; Function subst-term-list substitutes a term for a variable
; in a list of terms.

(defun subst-term-list (l v e)
  (declare (xargs :guard (and (wft-d-list l)
			      (variable-term v)
			      (wft-d e))))
  (if (atom l)
      nil ; I would like to change this to l, but it breaks not-member-vars...
      (cons (cond ((variable-term (car l))
		   (if (equal (car l) v)
		       e
		       (car l)))
		  ((domain-term (car l)) (car l))
		  ((atom (car l)) (car l))  ; shouldn't happen for wft-d
		  (t (cons (caar l) (subst-term-list (cdar l) v e))))
	    (subst-term-list (cdr l) v e))))

; Show that subst-term-list preserves well-formedness.

(defthm subst-term-list-wf
  (implies (and (wft-d-list l)
		(wft-d e))
	   (wft-d-list (subst-term-list l v e))))

; Function subst-free-var substitutes a term for free
; occurrences of a variable in a formula.

(defun subst-free-var (f v e)
  (declare (xargs :guard (and (wff-d f)
			      (variable-term v)
			      (wft-d e))))
  (cond ((wfnot f) (list 'not (subst-free-var (a1 f) v e)))
	((or (wfand f)
	     (wfor f)
	     (wfimp f)
	     (wfiff f)) (list (car f)
			      (subst-free-var (a1 f) v e) 
			      (subst-free-var (a2 f) v e)))
	((or (wfall f)
	     (wfexists f)) (if (equal (a1 f) v)
			       f
			       (list (car f)
				    (a1 f)
				    (subst-free-var (a2 f) v e))))
	((wfa-d f) (cons (car f) (subst-term-list (cdr f) v e)))
	(t f)))

; Prove that subst-free-var preserves well-formedness
; (This takes about 90 seconds if wff-is-wff-d is available.)

(defthm subst-free-var-wf
  (implies (and (wff-d f)
		(wft-d e))
	   (wff-d (subst-free-var f v e))))

; Function wff-count counts the number of nodes in a formula.
; It is useful for proving termination of functions (on formulas)
; that change atomic formulas.

(defun wff-count (f)
  (declare (xargs :guard (wff-d f)))
  (cond ((wfnot f) (+ 1 (wff-count (a1 f))))
	((or (wfand f)
	     (wfor f)
	     (wfimp f)
	     (wfiff f)) (+ 1 (wff-count (a1 f)) (wff-count (a2 f))))
	((or (wfall f)
	     (wfexists f)) (+ 1 (wff-count (a2 f))))
	(t 1)))

; Show that subst-free-var preserves the wff-count of a formula.

(defthm subst-free-var-preserves-wff-count
  (equal (wff-count (subst-free-var f v e))
	 (wff-count f))))

;===================== some set definitions and lemmas

(defun set-equal (a b)
  (declare (xargs :guard (and (true-listp a)
                              (true-listp b))))
  (and (subsetp-equal a b)
       (subsetp-equal b a)))

(defthm subset-cons
  (implies (subsetp-equal a b)
	   (subsetp-equal a (cons x b)))))

(defthm member-union-or
  (implies (or (member-equal x s1)
	       (member-equal x s2))
	   (member-equal x (union-equal s1 s2))))

(defthm subset-union
  (implies (and (subsetp-equal a b)
		(subsetp-equal c d))
	   (subsetp-equal (union-equal a c)
			  (union-equal b d))))
(defthm subset-reflexive
  (subsetp-equal x x))

(defthm set-equal-reflexive
  (set-equal x x))

(defun remove-equal (x l)
  (declare (xargs :guard (true-listp l)))
  (if (atom l)
      nil
      (if (equal x (car l))
	  (remove-equal x (cdr l))
	  (cons (car l) (remove-equal x (cdr l))))))

(defthm removed-element-is-not-member
  (not (member-equal x (remove-equal x s))))

(defthm not-member-union-equal
  (implies (and (not (member-equal x s1))
		(not (member-equal x s2)))
	   (not (member-equal x (union-equal s1 s2)))))

(defthm remove-union
  (implies
   (and (remove-equal x s1)
	(remove-equal x s2))
   (remove-equal x (union-equal s1 s2))))

(defthm subset-equal-remove
  (implies (subsetp-equal a b)
	   (subsetp-equal (remove-equal x a)
			  (remove-equal x b)))
  :hints (("Goal"
         :do-not generalize)))

(defthm not-member-not-member-remove
  (implies (not (member-equal y s))
	   (not (member-equal y (remove-equal x s)))))

(defthm union-equal-preserves-true-listp
  (implies (true-listp y)
	   (true-listp (union-equal x y))))

(defthm union-not-nil-if-either-argument-is-a-cons
  (implies (or (consp x) (consp y))
	   (union-equal x y)))

; end of set lemmas

;=================================================================
;
; In this section we extract sets of symbols from formulas.

; Domain terms.

(defun domain-terms-in-term-list (a)
  (declare (xargs :guard (wft-d-list a)))
  (if (atom a)
      nil
      (union-equal (cond ((domain-term (car a)) (list (car a)))
			 ((variable-term (car a)) nil)
			 ((atom (car a)) nil)  ; won't happen for wft-d-list
			 (t (domain-terms-in-term-list (cdar a))))
		   (domain-terms-in-term-list (cdr a)))))

(defun domain-terms-in-formula (f)
  (declare (xargs :guard (wff-d f)))
  (cond ((wfnot f) (domain-terms-in-formula (a1 f)))
	((or (wfand f)
	     (wfor f)
	     (wfimp f)
	     (wfiff f)) (union-equal (domain-terms-in-formula (a1 f))
				     (domain-terms-in-formula (a2 f))))
	((or (wfall f)
	     (wfexists f)) (domain-terms-in-formula (a2 f)))
	((wfa-d f) (domain-terms-in-term-list (cdr f)))
	(t nil)))

(defun domain-term-list (x)
  (declare (xargs :guard t))
  (and (true-listp x)
       (if (atom x)
	   t
	   (and (domain-term (car x))
		(domain-term-list (cdr x))))))

(defthm domain-terms-in-term-list-return
  (domain-term-list (domain-terms-in-term-list x)))

(defthm domain-terms-in-formula-return
  (domain-term-list (domain-terms-in-formula x)))

; Function symbols.  We extract a list of (function-symbol . arity) pairs.
; Recall that the function-symbol can be (symbolp . index).

(defun func-arity-in-term-list (a)
  (declare (xargs :guard (wft-d-list a)
		  :verify-guards nil))
  (if (atom a)
      nil
      (union-equal (cond ((domain-term (car a)) nil)
			 ((variable-term (car a)) nil)
			 ((atom (car a)) nil)  ; won't happen for wft-d-list
			 (t (union-equal (list (cons (caar a) (len (cdar a))))
					 (func-arity-in-term-list (cdar a)))))
		   (func-arity-in-term-list (cdr a)))))

(defthm true-listp-func-arity-in-term-list ; for verifying guards
  (true-listp (func-arity-in-term-list a)))

(verify-guards func-arity-in-term-list)

(defun func-arity-in-formula (f)
  (declare (xargs :guard (wff-d f)))
  (cond ((wfnot f) (func-arity-in-formula (a1 f)))
	((or (wfand f)
	     (wfor f)
	     (wfimp f)
	     (wfiff f)) (union-equal (func-arity-in-formula (a1 f))
				     (func-arity-in-formula (a2 f))))
	((or (wfall f)
	     (wfexists f)) (func-arity-in-formula (a2 f)))
	((wfa-d f) (func-arity-in-term-list (cdr f)))
	(t nil)))

(defun func-arity-list (x)
  (declare (xargs :guard t))
  (and (true-listp x)
       (if (atom x)
	   t
	   (and (consp (car x))
		(function-symbol (caar x))
		(natp (cdar x))
		(func-arity-list (cdr x))))))

(defthm union-equal-preserves-func-arity-list
  (implies (and (func-arity-list x)
		(func-arity-list y))
	   (func-arity-list (union-equal x y))))

(defthm func-arity-in-term-list-return
  (implies (wft-d-list l)
	   (func-arity-list (func-arity-in-term-list l))))

(defthm func-arity-in-formula-return
  (func-arity-list (func-arity-in-formula f)))

; Relation symbols.  We extract a list of (relation-symbol . arity) pairs.

(defun rel-arity-in-formula (f)
  (declare (xargs :guard (wff-d f)))
  (cond ((wfnot f) (rel-arity-in-formula (a1 f)))
	((or (wfand f)
	     (wfor f)
	     (wfimp f)
	     (wfiff f)) (union-equal (rel-arity-in-formula (a1 f))
				     (rel-arity-in-formula (a2 f))))
	((or (wfall f)
	     (wfexists f)) (rel-arity-in-formula (a2 f)))
	((wfa-d f) (list (cons (car f) (len (cdr f)))))
	(t nil)))

(defun rel-arity-list (x)
  (declare (xargs :guard t))
  (and (true-listp x)
       (if (atom x)
	   t
	   (and (consp (car x))
		(relation-symbol (caar x))
		(natp (cdar x))
		(rel-arity-list (cdr x))))))

(defthm rel-arity-in-formula-return
  (rel-arity-list (rel-arity-in-formula x)))

; Variables.  Extract the list of variables in a term.

(defun vars-in-term-list (l)
  (declare (xargs :guard (wft-d-list l)))
  (if (atom l)
      nil
      (union-equal (cond ((variable-term (car l)) (list (car l)))
			 ((domain-term (car l)) nil)
			 ((atom (car l)) nil)  ; won't happen for wft-d-list
			 (t (vars-in-term-list (cdar l))))
		   (vars-in-term-list (cdr l)))))

; ============================================== free variables
; Function free-vars gets the set of free variables in a formula.
; It uses union-equal, which is too complicated, I think.
; Originally, I had a function closed (wff var-list), which
; checks if every free variable in wff is a member of var-list.
; (See wff/first-order-q.)
; Since the main use of free-vars is to check if a formula is
; closed, function closed could be used instead.  (I don't remember
; exactly why I switched to free-vars.)  I think we should switch back.

(defun free-vars (f)
  (declare (xargs :guard (wff-d f)))
  (cond ((wfnot f) (free-vars (a1 f)))
	((or (wfand f)
	     (wfor f)
	     (wfimp f)
	     (wfiff f)) (union-equal (free-vars (a1 f)) (free-vars (a2 f))))
	((or (wfall f)
	     (wfexists f)) (remove-equal (a1 f) (free-vars (a2 f))))
	((wfa-d f) (vars-in-term-list (cdr f)))
	(t nil)))

(defthm free-vars-true-listp
  (true-listp (free-vars f)))

(defun var-list (x)
  (declare (xargs :guard t))
  (and (true-listp x)
       (if (atom x)
	   t
	   (and (variable-term (car x))
		(var-list (cdr x))))))

(defthm vars-in-term-list-return
  (var-list (vars-in-term-list l)))

(defthm free-vars-returns
  (var-list (free-vars f)))

; Here are a bunch of lemmas about free-vars.  These were written
; to get the guards verified for feval.  I put them here instead
; of with feval, because they might be useful elsewhere.

(defthm vars-in-term-list-parts
  (implies (and (wft-d-list l)
		(equal (vars-in-term-list l) nil))
	   (and (equal (vars-in-term-list (cdar l)) nil)
		(equal (vars-in-term-list (cdr l)) nil))))

(defthm not-member-vars-terms-not-member-subst-vars-terms
  (implies (and (domain-term e)
		(variable-term x)
		(wft-d-list l)
		(not (member-equal y (vars-in-term-list l))))
	   (not (member-equal y (vars-in-term-list
				 (subst-term-list l x e))))))

(defthm free-vars-terms-subset-free-vars-terms-subst
  (implies (and (domain-term e)
		(variable-term x)
		(wft-d-list l))
	   (subsetp-equal (vars-in-term-list (subst-term-list l x e))
			  (vars-in-term-list l))))

(defthm free-vars-subset-free-vars-subst
  (implies (and (domain-term e)
		(variable-term x))
	   (subsetp-equal (free-vars (subst-free-var f x e))
			  (free-vars f))))

(defthm not-member-x-subst-x-terms
  (implies (and (domain-term e)
                (variable-term x)
                (wft-d-list l))
           (not (member-equal x (vars-in-term-list
                                 (subst-term-list l x e)))))
  :hints (("Goal"
         :do-not generalize)))

(defthm not-member-x-subst-x
  (implies (and (domain-term e)
                (variable-term x))
           (not (member-equal x (free-vars (subst-free-var f x e))))))
