(in-package "ACL2")

; This file is mostly about interpretaions of first-order formulas.
;
; Author:  William McCune (mccune@mcs.anl.gov).

(include-book "/home/mccune/acl2/first-order/formulas-g")

; These next few functions do an exponential expansion
; of a list.  For example, (expand '(a b) 3) gives 
; the 8 (2^3) length 3 lists containing 'a and 'b.
; The result has the standard order, ((a a a) (a a b) ...),
; which is relevant in the definition of relations and functions
; below, which are defined in terms of (expand domain arity).

(include-book "/home/mccune/acl2/acl2-sources/books/arithmetic/top")

(defun cons-to-each (x tuples)
  (declare (xargs :guard t))
  (if (atom tuples)
      nil
      (cons (cons x (car tuples))
	    (cons-to-each x (cdr tuples)))))

(defun alphabet-times-tuples (alphabet tuples)
  (declare (xargs :guard t))
  (if (atom alphabet)
      nil
      (append (cons-to-each (car alphabet) tuples)
	      (alphabet-times-tuples (cdr alphabet) tuples))))

(defun expand (alphabet n)
  (declare (xargs :guard (natp n)))
  (if (zp n) 
      (list nil)
      (alphabet-times-tuples alphabet (expand alphabet (1- n)))))

; Define a predicate tuplep to check if a tuple is
; over a given alphabet and has a given arity.

(defun tuplep (tuple alphabet arity)
  (declare (xargs :guard (and (true-listp alphabet)
			      (natp arity))))
  (and (true-listp tuple)
       (equal (len tuple) arity)
       (subsetp-equal tuple alphabet)))

;==============================================
;
; In this section we prove that (expand alphabet n) contains
; exactly the correct tuples; that is, all of the length-n tuples
; over the alphabet.

(defthm correct-tuples-are-in-alphabet-times-tuples
  (implies (and (member-equal a alphabet)
		(member-equal w tuples))
	   (member-equal (cons a w) (alphabet-times-tuples alphabet tuples))))

(defthm tuplep-in-expansion
  (implies (tuplep tup alphabet arity)
	   (member-equal tup (expand alphabet arity))))

(defthm alphabet-times-tuples-contains-correct-tuples
  (implies (member-equal (cons a w) (alphabet-times-tuples alphabet tuples))
	   (and (member-equal a alphabet)
		(member-equal w tuples))))

; The following induction scheme recurses through a list and a nat together.

(defun tuplep-induct-scheme (tup arity)
  (declare (xargs :guard (natp arity)))
  (if (or (atom tup) (zp arity))
      nil
      (tuplep-induct-scheme (cdr tup) (1- arity))))

; (The following proof is interesting;  it does lots of generalization
; and 8 inductions.)

(defthm expansion-contains-tupleps
  (implies (and (natp arity)
		(member-equal tup (expand alphabet arity)))
	   (tuplep tup alphabet arity))
  :hints (("Goal"
	   :induct (tuplep-induct-scheme tup arity))))

;==============================================================
; Interpretations
; 
; An interpretation consists of
;  (1) a nonempty list representing the domain,
;  (2) a list of relations over the domain,
;  (3) a list of functions from the domain to the domain.
;
; The relations and functions are just alists taking tuples
; of domain elements to ('true 'false) for relations, and to
; domain elements for functions.
;
; I don't know what, if anything, this means for infinite domains.
;
; Domain of an interpretaion: a nonempty list of unique domain-terms.
; At first, I had this as a cons tree, because it must be nonempty;
; but I found that it's more convenient to have it as a list, because
; we have to a lot list-like operations on it.  I also considered
; specifying the domain with a natural n, and assuming the elements
; of the domain are 0,...,n-1.  This might be a better idea, but I
; didn't want to rely too much on arithmetic.

(defun domainp (x)
  (declare (xargs :guard t))
  (cond ((not (true-listp x))           nil)
	((atom x)                       nil)
	((not (domain-term (car x)))    nil)
	((member-equal (car x) (cdr x)) nil)
	((atom (cdr x))                 t)
	(t                              (domainp (cdr x)))))

; Function cdrs-in-range (a r) checks if every cdr of an alist
; is a member of r.

(defun cdrs-in-range (a r)
  (declare (xargs :guard (and (alistp a)
			      (true-listp r))))
  (if (atom a)
      t
      (and (member-equal (cdar a) r)
	   (cdrs-in-range (cdr a) r))))

(defun relationp (domain arity rel)
  (declare (xargs :guard (and (natp arity)
			      (true-listp domain))))
  (and (alistp rel)
       (equal (strip-cars rel) (expand domain arity))
       (cdrs-in-range rel '(true false))))

(defun funcp (domain arity func)
  (declare (xargs :guard (and (natp arity)
			      (true-listp domain))))
  (and (alistp func)
       (equal (strip-cars func) (expand domain arity))
       (cdrs-in-range func domain)))

; Here are examples of relations and functions.  Note that the
; tuples have to be in the standard order.
;
; (relationp '(0 1) 0 '((nil . true)))
;  
; (relationp '(0 1) 2 '(((0 0) . true)
;                       ((0 1) . false)
;                       ((1 0) . true)
;                       ((1 1) . true)))
;  
; (funcp '(0 1) 0 '((nil . 1)))
;  
; (funcp '(0 1) 2 '(((0 0) . 1)
;                   ((0 1) . 0)
;                   ((1 0) . 0)
;                   ((1 1) . 1)))

(defun relation-list (a domain)
  (declare (xargs :guard (domainp domain)))
  (cond ((not (alistp a)) nil)
	((null a) t)
	((not (consp (caar a))) nil)
	(t (let ((sym (caaar a)) (arity (cdaar a)) (rel (cdar a)))
		 (and (relation-symbol sym)
		      (natp arity)
		      (relationp domain arity rel)
		      (relation-list (cdr a) domain))))))

(defun function-list (a domain)
  (declare (xargs :guard (domainp domain)))
  (cond ((not (alistp a)) nil)
	((null a) t)
	((not (consp (caar a))) nil)
	(t (let ((sym (caaar a)) (arity (cdaar a)) (rel (cdar a)))
		 (and (function-symbol sym)
		      (natp arity)
		      (funcp domain arity rel)
		      (function-list (cdr a) domain))))))

; An interpretation consists of a domain and lists of relations
; and functions over the domain: (domainp . (relation-list . function-list))

(defun interp (i)
  (declare (xargs :guard t))
  (and (consp i)
       (domainp (car i))
       (consp (cdr i))
       (relation-list (cadr i) (car i))
       (function-list (cddr i) (car i))))

; The following macros are for interpretations.

(defmacro domain    (i) (list 'car  i))
(defmacro relations (i) (list 'cadr i))
(defmacro functions (i) (list 'cddr i))
  
(defun domain-tuplep (l)
  (declare (xargs :guard t))
  (and (true-listp l)
       (if (null l)
	   t
	   (and (domain-term (car l))
		(domain-tuplep (cdr l))))))

; Here is an example of an interpretation.  Note the nested alists: 
; first to get the function or relation, then second to get the
; value of the tuple.
; 
; (interp (cons '(0 1)
;               (cons '(((r . 0) (nil . true))
;                       ((p . 1) ((0) . true) ((1) . false)))
;                     '(((a . 0) (nil . 1))
;                       ((g . 1) ((0) . 1) ((1) . 0))))))


; Functions flookup and rlookup look up the value of a tuple
; in an interpretation.  If something goes wrong
; (including the case where the interpretaion doesn't interpret
; the symbol), nil is returned.  If all is well, flookup
; returns a domain-term, and rlookup returns 'true or 'false.
; (The guards should probably be strengthened: the interpretation
; should interpret all of the symbols, and the tuple should consist
; of domain-terms.)

(defthm func-nested-alist  ; This helps to verify guards for flookup below.
  (implies (and (function-list funcs domain)
		(assoc-equal x funcs))
	   (alistp (cdr (assoc-equal x funcs)))))

(defun flookup (sym tuple i)
  (declare (xargs :guard (and (function-symbol sym)
			      (true-listp tuple)
			      (interp i))))
  (let ((func (assoc-equal (cons sym (len tuple)) (cddr i))))
    (if (not (consp func))
	nil
        (let ((val (assoc-equal tuple (cdr func))))
	  (if (not (consp val))
	      nil
	      (cdr val))))))

(defthm rel-nested-alist  ; This helps to verify guards for rlookup below.
  (implies (and (relation-list rels domain)
		(assoc-equal x rels))
	   (alistp (cdr (assoc-equal x rels)))))

(defun rlookup (sym tuple i)
  (declare (xargs :guard (and (relation-symbol sym)
			      (true-listp tuple)
			      (interp i))))

  (let ((rel (assoc-equal (cons sym (len tuple)) (cadr i))))
    (if (not (consp rel))
	nil
        (let ((val (assoc-equal tuple (cdr rel))))
	  (if (not (consp val))
	      nil
	      (cdr val))))))

(defun eval-term-list (l i)
  (declare (xargs :guard (and (wft-d-list l)
			      (equal (vars-in-term-list l) nil)
			      (interp i))))
  (if (atom l)
      nil
      (cons (cond ((domain-term (car l)) (car l))
		  ((atom (car l)) nil) ; won't happen for wft-d-list
		  (t (flookup (caar l) (eval-term-list (cdar l) i) i)))
	    (eval-term-list (cdr l) i))))

(defun eval-atomic-formula (a i)
  (declare (xargs :guard (and (wfa-d a)
			      (equal (free-vars a) nil)
			      (interp i))))
  (rlookup (car a) (eval-term-list (cdr a) i) i))

;========================
; For verifying guards of feval below.  This is messy, because feval
; has a guard that the formula is closed, and we have to prove that
; substituting a domain term keeps it closed.

(defthm special-set-rule-for-nasty-lemma
  (implies (and (equal (remove-equal x a) nil)
		(true-listp b)
		(subsetp-equal b a)
		(not (member-equal x b)))
	   (equal b nil))
  :hints (("Goal"
         :induct (remove-equal x a)))
  :rule-classes nil)

(defthm nasty-lemma-to-get-feval-guards-verified
  (implies (and (domain-term e)
		(variable-term x)
		(wff-d f)
		(not (remove-equal x (free-vars f))))
	   (not (free-vars (subst-free-var f x e))))
  :hints (("Goal"
	   :do-not-induct t
	   :use ((:instance special-set-rule-for-nasty-lemma
			    (a (free-vars f))
			    (b (free-vars (subst-free-var f x e)))
			    (x x))))))

; Function feval evaluates a formula in an interpretation.
; There is a simulated mutual recursion: (1) through the formula
; structure, and (2) for quantified formulas, through the domain
; of the interpretation.  Return t or nil.  Note that 
; eval-atomic-formula returns 'true or 'false (or nil on error).
;
; This should probably be changed so that if anything goes
; wrong during the evaluation, nil is returned, and if all goes
; well, 'true or 'false is returned.

(defun feval (p dom i)
  (declare (xargs :measure (cons (wff-count p) (acl2-count dom))
		  :guard (and (wff-d p)	
			      (equal (free-vars p) nil)
			      (domainp dom)
			      (interp i))))
  (cond ((equal p 'true)    t)
	((equal p 'false)   nil)

	((wfand p) (and     (feval (a1 p) dom i) (feval (a2 p) dom i)))
	((wfor  p) (or      (feval (a1 p) dom i) (feval (a2 p) dom i)))
	((wfnot p) (not     (feval (a1 p) dom i)))
	((wfimp p) (implies (feval (a1 p) dom i) (feval (a2 p) dom i)))
	((wfiff p) (equal   (feval (a1 p) dom i) (feval (a2 p) dom i)))

	; For 'all and 'exists, we don't want to recurse to the end
	; of the domain, because domains are nonempty.  This was
	; cleaner when domains were cons trees.

	((wfall p) (and (feval (subst-free-var (a2 p) (a1 p) (car dom))
			       (car i) i)
			(if (atom (cdr dom))
			    t
			    (feval p (cdr dom) i))))

	((wfexists p) (or (feval (subst-free-var (a2 p) (a1 p) (car dom))
				 (car i) i)
			  (if (atom (cdr dom))
			      nil
			      (feval p (cdr dom) i))))

	(t (equal (eval-atomic-formula p i) 'true))))

; Here are some tests of feval.

(defthm all-x-and-p-q
  (implies (and (domainp dom)
		(variable-term x))
	   (equal (feval (list 'all x (list 'and p q)) dom i)
		  (feval (list 'and
			       (list 'all x p)
			       (list 'all x q)) dom i)))
  :hints (("Goal"
	   :induct (domainp dom))))

(defthm all-not-not-exists
  (implies (and (domainp dom)
		(variable-term x))
	   (equal (feval (list 'all x (list 'not p)) dom i)
		  (feval (list 'not (list 'exists x p)) dom i)
		  ))
  :hints (("Goal"
	   :induct (domainp dom)
	   )))

