(in-package "ACL2")

; This file is about normal forms of first-order formulas.
;
; Author:  William McCune (mccune@mcs.anl.gov).

(include-book "/home/mccune/acl2/first-order/interpretations-g")

; Normal forms.  For each,
;  1.  define a function nf to rewrite to it,
;  2.  prove that nf preserves well-formedness,
;  3.  define a recognizer nfp,
;  4.  prove that nf puts formulas in nfp,
;  5.  soundness: prove that nf preserves evaluation,
;  6.  prove that nf doesn't change the set of free variables.

;==================================================================
; Function nnf converts a formula to negation normal form.
; That is, in terms of and/or/not, where all nots are up against
; simple formulas.  ('true and 'false are not simplified away.)
	 
(defun nnf (f)
  (declare (xargs :guard (wff-d f)))
  (cond ((wfand f) (list 'and (nnf (a1 f)) (nnf (a2 f))))
	((wfor  f) (list 'or  (nnf (a1 f)) (nnf (a2 f))))
	((wfimp f) (list 'or  (nnf (list 'not (a1 f))) (nnf (a2 f))))
	((wfiff f) (list 'and 
			 (list 'or (nnf (list 'not (a1 f))) (nnf (a2 f)))
			 (list 'or (nnf (a1 f)) (nnf (list 'not (a2 f))))))
	((wfall    f) (list 'all    (a1 f) (nnf (a2 f))))
	((wfexists f) (list 'exists (a1 f) (nnf (a2 f))))
	((wfnot f)
	 (cond ((wfand (a1 f)) (list 'or
				     (nnf (list 'not (a1 (a1 f))))
				     (nnf (list 'not (a2 (a1 f))))))
	       ((wfor (a1 f)) (list 'and 
				    (nnf (list 'not (a1 (a1 f))))
				    (nnf (list 'not (a2 (a1 f))))))
	       ((wfimp (a1 f)) (list 'and 
				     (nnf (a1 (a1 f)))
				     (nnf (list 'not (a2 (a1 f))))))
	       ((wfiff (a1 f)) (list 'and 
				     (list 'or 
					   (nnf (a1 (a1 f)))
					   (nnf (a2 (a1 f))))
				     (list 'or 
					   (nnf (list 'not (a1 (a1 f))))
					   (nnf (list 'not (a2 (a1 f)))))))
	       ((wfall (a1 f))
		(list 'exists (a1 (a1 f)) (nnf (list 'not (a2 (a1 f))))))
	       ((wfexists (a1 f))
		(list 'all (a1 (a1 f)) (nnf (list 'not (a2 (a1 f))))))
	       ((wfnot (a1 f)) (nnf (a1 (a1 f))))
	       (t f)))
	(t f)))

; Prove that nnf preserves well-formedness.

(defthm nnf-wff-d    ; 547 seconds if wff-is-wff-d is available
  (implies (wff-d f)
	   (wff-d (nnf f))))

;  :hints (("Goal"
;	   :in-theory (disable wff-is-wff-d)
;	   )))

; Function nnfp checks if a formula is in negation normal form.
; Should it be defined positively (by what it must be like) or
; negatively (what is not allowed)?  For wff-d input, the result
; is the same.  Let's do it positively.

(defun nnfp (f)
  (declare (xargs :guard t))
  (cond ((wfnot f) (simple-formula-d (a1 f)))
	((wfand f) (and (nnfp (a1 f)) (nnfp (a2 f))))
	((wfor  f) (and (nnfp (a1 f)) (nnfp (a2 f))))
	((wfall    f) (nnfp (a2 f)))
	((wfexists f) (nnfp (a2 f)))
	((simple-formula-d f) t)
	(t nil)))

; Prove nnf applied to a wff gives negation normal form.

(defthm nnf-nnfp
  (implies (wff-d x)
	   (nnfp (nnf x))))

; Prove that subst-free-var preserves nnfp.

(defthm subst-free-var-nnf
  (implies (and (nnfp f)
                (wft-d e))
           (nnfp (subst-free-var f v e))))

; Prove that nnf is sound.

(defthm subst-nnf-commute
  (equal (subst-free-var (nnf f) x e)
	 (nnf (subst-free-var f x e))))

; Function feval-i is used only for setting up inductions.  I couldn't
; make some inductions on feval work, because for the 'and, 'or,
; 'all, and 'exists cases, (and possibly others), it was giving cases
; that seemed complicated and unnecessary.
; For example, the wfand case gave two subcases, I guess because
; (and a b) is really (if a b nil).

(defun feval-i (f dom i)
  (declare (xargs :measure (cons (wff-count f) (acl2-count dom))
		  :guard (and (wff-d f) (domainp dom) (interp i))))
  (cond ((equal f 'true)     nil)
	((equal f 'false)    nil)
	((wfnot f) (feval-i (a1 f) dom i))
	((wfand f) (cons (feval-i (a1 f) dom i) (feval-i (a2 f) dom i)))
	((wfor  f) (cons (feval-i (a1 f) dom i) (feval-i (a2 f) dom i)))
	((wfimp f) (cons (feval-i (a1 f) dom i) (feval-i (a2 f) dom i)))
	((wfiff f) (cons (feval-i (a1 f) dom i) (feval-i (a2 f) dom i)))

	((wfall f) (if (atom (cdr dom))
		       (feval-i (subst-free-var (a2 f) (a1 f) (car dom))
				(car i) i)
		       (cons (feval-i (subst-free-var (a2 f) (a1 f) (car dom))
				      (car i) i)
			     (feval-i f (cdr dom) i))))
	((wfexists f) (if (atom (cdr dom))
			  (feval-i (subst-free-var (a2 f) (a1 f) (car dom))
				   (car i) i)
			  (cons (feval-i (subst-free-var (a2 f) (a1 f)
							 (car dom))
					 (car i) i)
				(feval-i f (cdr dom) i))))
	(t nil)))


(defthm nnf-sound-for-not
  (implies (and (interp i)
		(domainp dom))
	   (equal (feval (nnf (list 'not f)) dom i)
		  (feval (list 'not (nnf f)) dom i)))
  :hints (("Goal"
	   :induct (feval-i f dom i))))

(defthm nnf-sound
  (implies (and (interp i)
		(domainp dom))
	   (equal (feval (nnf f) dom i)
		  (feval f dom i)))
  :hints (("Goal"
	   :do-not generalize
	   :induct (feval-i f dom i))))

; Prove that nnf preserves the set of free variables.

(defthm subset-union-2                     ; move to set book
  (implies (subsetp-equal a b)
	   (equal (union-equal a b) b)))

(defthm union-equal-idempotent             ; move to set book
  (equal (union-equal x x) x))

(defthm nnf-preserves-free-vars
  (equal (free-vars (nnf f)) (free-vars f)))

; ============================= CNF - conjunctive normal form
; 

(defun dist-or-and-2 (p q)
  (declare (xargs :guard (and (wff-d p) (wff-d q))))
  (if (wfand q)
      (list 'and (dist-or-and-2 p (a1 q)) (dist-or-and-2 p (a2 q)))
      (list 'or p q)))

(defun dist-or-and (p q)
  (declare (xargs :guard (and (wff-d p) (wff-d q))))
  (if (wfand p)
      (list 'and (dist-or-and (a1 p) q) (dist-or-and (a2 p) q))
      (dist-or-and-2 p q)))

(defthm dist-or-and-2-wff-d  ; help verify guards for cnf below
  (implies (and (wff-d p)
		(wff-d q))
	   (wff-d (dist-or-and-2 p q))))

(defthm dist-or-and-wff-d  ; help verify guards for cnf below
  (implies (and (wff-d p)
		(wff-d q))
	   (wff-d (dist-or-and p q))))

(defun cnf (f) 
  (declare (xargs :guard (wff-d f)))
  (cond ((wfand f) (list 'and (cnf (a1 f)) (cnf (a2 f))))
	((wfor  f) (dist-or-and (cnf (a1 f)) (cnf (a2 f))))
	((wfall    f) (list 'all    (a1 f) (cnf (a2 f))))
	((wfexists f) (list 'exists (a1 f) (cnf (a2 f))))
	(t f)))

; Prove that cnf preserves well-formedness.

(defthm cnf-wff-d
  (implies (wff-d f)
	   (wff-d (cnf f))))

; Function cnfp checks if a formula is in conjunctive normal form.

(defun cnfp (f)
  (declare (xargs :guard t))
  (cond ((simple-formula-d f) t)
	((wfor f) (and (cnfp (a1 f))
		       (cnfp (a2 f))
		       (not (wfand (a1 f)))
		       (not (wfand (a2 f)))))
	((wfand f) (and (cnfp (a1 f))
			(cnfp (a2 f))))
	((wfnot f) (simple-formula-d (a1 f)))
	((wfall    f) (cnfp (a2 f)))
	((wfexists f) (cnfp (a2 f)))
	(t nil)))

; Prove that cnf formulas are always nnf.

(defthm cnfp-nnfp
  (implies (cnfp f) (nnfp f)))

; Prove that cnf rewrites an nnf formula into conjunctive normal form.

(defthm dist-or-and-2-cnfp
  (implies (and (cnfp p)
		(cnfp q)
		(not (wfand p)))
	   (cnfp (dist-or-and-2 p q))))

(defthm dist-or-and-cnfp
  (implies (and (cnfp p)
		(cnfp q))
	   (cnfp (dist-or-and p q))))

(defthm cnf-cnfp
  (implies (nnfp f)
	   (cnfp (cnf f))))

; Prove that subst-free-var preserves cnfp.

(defthm subst-free-var-preserves-car
  (equal (car (subst-free-var f v e)) (car f)))

(defthm subst-free-var-cnf
  (implies (and (cnfp f)
                (wft-d e))
           (cnfp (subst-free-var f v e))))

;--------------------------------- soundness of CNF

(defthm subst-dist-dist-2
  (equal (subst-free-var (dist-or-and-2 p q) x e)
	 (dist-or-and-2 (subst-free-var p x e)
			(subst-free-var q x e))))

(defthm subst-dist-dist
  (equal (subst-free-var (dist-or-and p q) x e)
	 (dist-or-and (subst-free-var p x e)
		      (subst-free-var q x e))))

(defthm subst-cnf-commute
  (equal (subst-free-var (cnf f) x e)
	 (cnf (subst-free-var f x e))))

(defthm dist-or-and-2-sound
  (equal (feval (dist-or-and-2 p q) dom i)
	 (feval (list 'or p q) dom i)))

(defthm dist-or-and-sound
  (equal (feval (dist-or-and p q) dom i)
	 (feval (list 'or p q) dom i))))

; Prove that function cnf is sound.

(defthm cnf-sound
  (implies (and (domainp dom)
		(interp i))
	   (equal (feval (cnf p) dom i)
		  (feval p dom i)))
  :hints (("Goal"
	   :do-not generalize
	   :induct (feval-i p dom i))))

; Prove that cnf preserves the set of free variables.
; 
; This section isn't working yet, because the set operations are messy.
;
; ;------------------------
; 
; (defaxiom special-set-union-rule-1
;   (implies (and (set-equal a (union-equal b c))
;                 (set-equal d (union-equal b e)))
;            (set-equal (union-equal a d)
;                       (union-equal b (union-equal c e)))))
;                  
; ;------------------------
; 
; (defthm dist-or-and-2-preserves-free-vars ; ok
;   (set-equal (free-vars (dist-or-and-2 f1 f2))
;              (free-vars (list 'or f1 f2)))
;   :hints (("Goal"
;            :in-theory (disable set-equal))))
; 
; (defthm dist-or-and-preserves-free-vars  ; needs work
;    (set-equal (free-vars (dist-or-and f1 f2))
;               (free-vars (list 'or f1 f2)))
;   :hints (("Goal"
;            :in-theory (disable set-equal))))
; 
; (defthm cnf-preserves-free-vars
;            (set-equal (free-vars (cnf f))
;                       (free-vars f)))
; 
