(in-package "ACL2")

; This file is mostly about renaming bound variables in first-order formulas.
;
; Author:  William McCune (mccune@mcs.anl.gov).

(include-book "/home/mccune/acl2/first-order/interpretations-g")

; Function safe-var-index returns a natp, say n, that does not
; occur as the variable index of any variable (bound or free)
; in the given formula.  The natp n can then be used to constuct
; a fresh variable (v . n) for any symbolp v.  This is useful
; for renaming bound variables.

(defun safe-var-index-term-list (l)
  (declare (xargs :guard (wft-d-list l)))
  (if (atom l)
      0
      (max (cond ((variable-term (car l)) (if (consp (car l))
					      (1+ (cdar l))
					      0))
		 ((domain-term (car l)) 0)
		 ((atom (car l)) 0)  ; won't happen for wft-d-list
		 (t (safe-var-index-term-list (cdar l))))
	   (safe-var-index-term-list (cdr l)))))

(defun safe-var-index (f)
  (declare (xargs :guard (wff-d f)))
  (cond ((wfnot f) (safe-var-index (a1 f)))
	((or (wfand f)
	     (wfor f)
	     (wfimp f)
	     (wfiff f)) (max (safe-var-index (a1 f))
			     (safe-var-index (a2 f))))
	((or (wfall f)
	     (wfexists f)) (max (if (consp (a1 f))
				    (1+ (cdr (a1 f)))
				    0)
				(safe-var-index (a2 f))))
	((wfa-d f) (safe-var-index-term-list (cdr f)))
	(t 0)))

; Function bound-occurrence (x f) checks if formula f contains
; a quantifier that binds variable x.

(defun bound-occurrence (x f)
  (declare (xargs :guard (and (variable-term x) (wff-d f))))
  (cond ((wfnot f) (bound-occurrence x (a1 f)))
	((or (wfand f)
	     (wfor f)
	     (wfimp f)
	     (wfiff f)) (or (bound-occurrence x (a1 f))
			    (bound-occurrence x (a2 f))))
	((or (wfall f)
	     (wfexists f)) (if (equal (a1 f) x)
			       t
			       (bound-occurrence x (a2 f))))
	(t nil)))

; Function var-occurrence-term-list (x l) checks if term-list l
; contains variable x.

(defun var-occurrence-term-list (x l)
  (declare (xargs :guard (and (variable-term x) (wft-d-list l))))
  (if (atom l)
      nil
      (or (cond ((variable-term (car l)) (equal (car l) x))
		((domain-term (car l)) nil)
		((atom (car l)) nil)  ; won't happen for wft-d-list
		(t (var-occurrence-term-list x (cdar l))))
	  (var-occurrence-term-list x (cdr l)))))
  
; Function free-occurrence (x f) checks if formula f contains
; x as a free varible.

(defun free-occurrence (x f)
  (declare (xargs :guard (and (variable-term x) (wff-d f))))
  (cond ((wfnot f) (free-occurrence x (a1 f)))
	((or (wfand f)
	     (wfor f)
	     (wfimp f)
	     (wfiff f)) (or (free-occurrence x (a1 f))
			    (free-occurrence x (a2 f))))
	((or (wfall f)
	     (wfexists f)) (if (equal x (a1 f))
			       nil
			       (free-occurrence x (a2 f))))
	((wfa-d f) (var-occurrence-term-list x (cdr f)))
	(t nil)))

; Prove that the safe-var-index routines return naturals.

(defthm safe-var-index-term-list-returns-natp
  (natp (safe-var-index-term-list l)))

(defthm safe-var-index-returns-natp
  (natp (safe-var-index l)))

; Here is a case where it's easier to prove theorems more
; general than we really need.

(defthm safe-ok-for-term-list-less
  (implies (and (natp n)
		(<= (safe-var-index-term-list l) n))
	   (not (var-occurrence-term-list (cons v n) l))))

(defthm safe-ok-for-free-less
  (implies (and (natp n)
		(<= (safe-var-index f) n))
	   (not (free-occurrence (cons v n) f))))

(defthm safe-ok-for-bound-less
  (implies (and (natp n)
		(<= (safe-var-index f) n))
	   (not (bound-occurrence (cons v n) f))))

; Here are the theorems we need---that safe-var-index works.
; These are simple corrolaries of the preceding theorems.

(defthm safe-ok-for-term-list
  (not (var-occurrence-term-list (cons v (safe-var-index-term-list l)) l)))

(defthm safe-ok-for-free
  (not (free-occurrence (cons v (safe-var-index f)) f)))

(defthm safe-ok-for-bound
  (not (bound-occurrence (cons v (safe-var-index f)) f)))

;================================================
;
; This is easily proved, but it slows things.
; (defthm not-free-not-change
;   (implies (not (free-occurrence x f))
;            (equal (subst-free-var f x e) f)))

(defthm subst-subst-free-var  ; this takes about 270 seconds.
  (implies (and (variable-term x)
		(variable-term y)
	        (not (bound-occurrence y f))
		(not (free-occurrence y f)))
	   (equal (subst-free-var (subst-free-var f x y) y tm)
		  (subst-free-var f x tm))))

;===================== step-wise rename

; Function rename-bound (f old new) renames the first bound occurrence
; of old to new.  Safeness of "new" is not checked.

(defun rename-bound (f old new)
  (declare (xargs :guard (and (wff-d f)
			      (variable-term old)
			      (variable-term new))))
  (cond ((wfnot f) (rename-bound (a1 f) old new))
	((or (wfand f)
	     (wfor f)
	     (wfimp f)
	     (wfiff f)) (if (bound-occurrence old (a1 f))
			    (list (car f)
				  (rename-bound (a1 f) old new)
				  (a2 f))
			    (list (car f)
				  (a1 f)
				  (rename-bound (a2 f) old new))))
	((or (wfall f)
	     (wfexists f)) (if (equal (a1 f) old)
			       (list (car f)
				     new
				     (subst-free-var (a2 f) (a1 f) new))
			       (list (car f)
				     (a1 f)
				     (rename-bound (a2 f) old new))))
	(t f)))


(defthm rename-bound-wff-d
  (implies (and (wff-d f)
		(variable-term new))
	   (wff-d (rename-bound f old new))))

;------------------------------------------

; This should be moved to interpretations-?

(defun domain-induct (dom)
  (declare (xargs :guard (domainp dom)))
  (if (atom (cdr dom))
      nil
      (domain-induct (cdr dom)))))

;----------------------------

(defthm rename-bound-sound-top-quant-equal ; ok
  (implies (and (wff-d f)
		(or (wfall f) (wfexists f))
		(equal (a1 f) x)
		(variable-term y)
	        (not (bound-occurrence y f))
		(not (free-occurrence y f))
		(domainp dom)
		(interp i)
		)
	   (equal (feval (rename-bound f x y) dom i)
		  (feval f dom i)))
  :hints (("Goal"
	   :induct (domain-induct dom))))

; Working on this, which should be true.
; 
; (defthm rename-bound-sound
;   (implies (and (wff-d f)
; 		(variable-term x)
; 		(variable-term y)
; 	        (not (bound-occurrence y f))
; 		(not (free-occurrence y f))
; 		(domainp dom)
; 		(interp i)
; 		)
; 	   (equal (feval (rename-bound f x y) dom i)
; 		  (feval f dom i))))

;===============================================
; Function quantified-vars (f) gets the list of variable occurrences
; that are bound by 'all or 'exists.

(defun quantified-vars (f)
  (declare (xargs :guard (wff-d f)))
  (cond ((wfnot f) (quantified-vars (a1 f)))
	((or (wfand f)
	     (wfor f)
	     (wfimp f)
	     (wfiff f)) (append (quantified-vars (a1 f))
				(quantified-vars (a2 f))))
	((or (wfall f)
	     (wfexists f)) (cons (a1 f) (quantified-vars (a2 f))))
	(t nil)))

(defun totren-3 (f qvars)
  (declare (xargs :guard (and (wff-d f)
			      (var-list qvars))))
  (if (atom qvars)
      f
      (totren-3 (rename-bound f (car qvars) (cons 'v (safe-var-index f)))
		(cdr qvars))))

; This function renames all bound variables to unique variable names
; of the form (cons 'v n) for some natural n.  The renaming goes
; left to right, incrementing n, starting with 1 more than the
; greatest index in f.
; 

(defun rename-all-bound-vars (f)
  (declare (xargs :guard (wff-d f)))
  (totren-3 f (quantified-vars f)))

; Now, we need to prove that rename-all-bound-vars does its job.
